/* SegorKat - A viewer for the electronic catalog of  
*            Segor electronics (www.segor.de). 
* 
* http://www.mydarc.de/dl9obn/programming/cpp/segorkat 
* 
* Copyright (C) 2005 by Dirk Baechle (dl9obn@darc.de) 
* 
* This program is free software; you can redistribute it and/or 
* modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation; either version 2 
* of the License, or (at your option) any later version. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public 
* License along with this program; if not, write to the 
* 
* Free Software Foundation, Inc. 
* 675 Mass Ave 
* Cambridge 
* MA 02139 
* USA 
* 
*/

/** 
\file skArticleOrderDialog.cpp 
\author Dirk Baechle 
\date 2005-07-02 
\version 1.0 
*/

/*------------------------------------------------------------- Includes */

#include "skArticleOrderDialog.h"

#include <qlayout.h>
#include <qpushbutton.h>
#include <q3textedit.h>
#include <qlineedit.h>
#include <q3hbox.h>
#include <q3groupbox.h>
#include <qpainter.h>
#include <qprinter.h>
#include <q3paintdevicemetrics.h>
#include <q3simplerichtext.h>
#include <qrect.h>
//Added by qt3to4:
#include <Q3VBoxLayout>

/*------------------------------------------------------ Class functions */

/** The constructor for the modal \a skArticleOrderDialog class. 
*/
skArticleOrderDialog::skArticleOrderDialog(const skSegorArticle &article,
                                         QWidget *parent, 
                                         QString name )
    : QDialog( parent, name, true)
    
{
    currentArticle = article;
    Q3BoxLayout *vb = new Q3VBoxLayout( this );

    Q3GroupBox *gbArticleInfo = new Q3GroupBox(1, Qt::Horizontal, tr("Artikel"), this);
    teArticleInfo = new Q3TextEdit( gbArticleInfo, "teArticleInfo" );
    teArticleInfo->setReadOnly(true);

    Q3GroupBox *gbOrderedUnits = new Q3GroupBox(1, Qt::Horizontal, tr("Bestellte Einheiten"), this);
    leOrderedUnits = new QLineEdit( gbOrderedUnits, "leOrderedUnits");

    // Is this article already ordered?
    if (0 == article.OrderedUnits)
    {
      // No, so try to set the minimum
      if (article.plPrices.size() > 0)
      {
        leOrderedUnits->setText(QString::number(article.plPrices[0].Units));
        currentArticle.OrderedUnits = article.plPrices[0].Units;
      }
      else
      {
        // No price information, so this article can not be
        // ordered...
        leOrderedUnits->setText(tr("Keine Preisinformationen!"));
        leOrderedUnits->setReadOnly(true);
        currentArticle.OrderedUnits = 0;
      }
    }
    else
    {
      leOrderedUnits->setText(QString::number(article.OrderedUnits));
    }

    Q3HBox *hbButtons = new Q3HBox( this );
    QPushButton *cancel = new QPushButton( tr("Abbruch"), hbButtons, "pbCancel" );
    QPushButton *ok = new QPushButton( tr("OK"), hbButtons, "pbOK" );
    connect(cancel, SIGNAL(clicked()), SLOT(reject()) );
    connect(ok, SIGNAL(clicked()), SLOT(accept()) );
    connect(leOrderedUnits, SIGNAL( textChanged(const QString &) ),
            this, SLOT( slotOrderedUnitsChanged(const QString &) ));

    vb->addWidget(gbArticleInfo);
    vb->addWidget(gbOrderedUnits);
    vb->addWidget(hbButtons);
    vb->activate();

    cancel->setFixedSize(cancel->width(), cancel->height());
    ok->setFixedSize(ok->width(), ok->height());

    resize(550, 400);

    QString text;
    getInfoText(currentArticle, text);
    teArticleInfo->setText(text);
}

/** Returns an info string \a s that contains some basic data of
* \a article in RichText/HTML format.
*/
void skArticleOrderDialog::getInfoText(const skSegorArticle &article,
                                      QString &s)
{
  s = "<QT><CENTER><TABLE>\n";
  
  // Name
  s += "<TR><TD><B>";
  s += tr("Name") + "</B>:</TD><TD>";
  s += QString(article.Name.c_str());
  s += "</TD></TR>\n";
  
  // Description
  s += "<TR><TD><B>";
  s += tr("Beschreibung") + "</B>:</TD><TD>";
  s += QString(article.Description.c_str());
  s += "</TD></TR>\n";
 
  // Order unit
  s += "<TR><TD><B>";
  s += tr("1 VPE enthlt") + "</B>:</TD><TD>";
  s += QString::number(article.PerUnit);
  s += QString(" ") + tr("Stck");
  s += "</TD></TR>\n";

  s += "</TABLE></CENTER></QT>\n";
}

/** Returns the const reference to the currently selected
* article.
*/
const skSegorArticle &skArticleOrderDialog::getArticle()
{
  return currentArticle;
}

/*--------------------------------------------------------------- Slots */


/** Prints the basic infos about the current article.
*/
void skArticleOrderDialog::slotPrint()
{
  QPrinter printer;
  printer.setFullPage(TRUE);
  if ( printer.setup( this ) ) 
  {
    QPainter p( &printer );
    Q3PaintDeviceMetrics metrics(p.device());
    int dpix = metrics.logicalDpiX();
    int dpiy = metrics.logicalDpiY();
    const int margin = 72; // pt
    QRect body(margin*dpix/72, margin*dpiy/72,
               metrics.width()-margin*dpix/72*2,
               metrics.height()-margin*dpiy/72*2 );

    QString rText;
    // Create rich text
    getInfoText(currentArticle, rText);

    Q3SimpleRichText richText( rText, QFont());
    richText.setWidth( &p, body.width() );
    QRect view( body );
    int page = 1;
    do 
    {
      richText.draw( &p, body.left(), body.top(), view, colorGroup() );
      view.moveBy( 0, body.height() );
      p.translate( 0 , -body.height() );
      p.drawText( view.right() - p.fontMetrics().width( QString::number(page) ),
                  view.bottom() + p.fontMetrics().ascent() + 5, QString::number(page) );
      if ( view.top()  >= richText.height() )
          break;
      printer.newPage();
      page++;
    } while (TRUE);
  }

}

/** Sets the new number of ordered units for the
* current article.
*/
void skArticleOrderDialog::slotOrderedUnitsChanged(const QString &orderedUnits)
{
  int orderedNew;
  bool success;

  orderedNew = orderedUnits.toInt(&success);

  if (true == success)
  {
    // Check against the minimum
    if (currentArticle.plPrices.size() > 0)
    {
      if (orderedNew >= currentArticle.plPrices[0].Units)
        currentArticle.OrderedUnits = orderedNew;
      
      leOrderedUnits->setText(QString::number(currentArticle.OrderedUnits));
    }
    else
    {
      // No price information, so this article can not be
      // ordered...
      leOrderedUnits->setText(tr("Keine Preisinformationen!"));
      leOrderedUnits->setReadOnly(true);
    }
  }
}

