/* SegorKat - A viewer for the electronic catalog of  
*            Segor electronics (www.segor.de). 
* 
* http://www.mydarc.de/dl9obn/programming/cpp/segorkat 
* 
* Copyright (C) 2005 by Dirk Baechle (dl9obn@darc.de) 
* 
* This program is free software; you can redistribute it and/or 
* modify it under the terms of the GNU General Public License 
* as published by the Free Software Foundation; either version 2 
* of the License, or (at your option) any later version. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public 
* License along with this program; if not, write to the 
* 
* Free Software Foundation, Inc. 
* 675 Mass Ave 
* Cambridge 
* MA 02139 
* USA 
* 
*/

/** 
\file skUncompress.cpp 
\author Dirk Baechle 
\date 2005-07-02 
\version 1.0 
*/

/*------------------------------------------------------------- Includes */

#ifdef SEGORKAT_QT
  #include <qapplication.h>
  #include <qstring.h>
  #include <q3textstream.h>
  #include <qfile.h>
  #include <q3progressdialog.h>
#endif

#include "skUncompress.h"
#include "skSegorFile.h"
#include "skHuffNode.h"
#include "skBitParser.h"
#include "skFileIO.h"
#include "skGlobal.h"

#include <assert.h>

#include <fstream>
#include <vector>

using namespace std;

/*-------------------------------------------------------------- Const values */

/** Size of the tree for Huffman decompression */
const unsigned int TREESIZE = 316;

/** Length of the LZSS buffers */
const unsigned int LZSSBUFLEN = 4096;

/** Number of files, contained in the Segor archive */
const int NUMFILES = 13;

/*------------------------------------------------------------ Functions */

/** Extracts a single file from the Segor archive, using
* run length encoding (RLE) and writes the output to the file.
* specified in \a fInfo.
* @param f File stream for the archive
* @param destDir The directory where the file is extracted to
* @param fInfo Info parameters for the file to be decompressed
*/
int rleExtractToFile(ifstream &f, 
                     string destDir,
                     const skSegorFile &fInfo)
{
  // Construct output filename
  string fileName(destDir);
  fileName += fInfo.Name;
  if (fInfo.Extension.length() > 0)
  {
    fileName += ".";
    fileName += fInfo.Extension;
  }

  ofstream fOut(fileName.c_str());

  if (fOut)
  {
    unsigned long int cnt = 0;
    unsigned char c, r, i;
    while ((!f.eof()) && (cnt < fInfo.Size))
    {
      c = readUChar(f);
      if (0x90 != c)
      {
        // Normal char -> store and write to output
        r = c;
        writeUChar(fOut, c);
      }
      else
      {
        // 0x90 encountered -> read next char
        c = readUChar(f);
        ++cnt;
        if (0x90 != c)
        {
          // Repeat last character c times
          for (i = 0; i < c; ++i)
            writeUChar(fOut, r);
        }
        else
        {
          // Convert special sequence 0x9090 to a single letter
          r = 0x90;
          writeUChar(fOut, r);
        }
      }
      ++cnt;
    }

    fOut.close();
    if (cnt == fInfo.Size)
      return OK;
  }

  return ERROR;
}

/** Extracts a single file from the Segor archive using
* Huffman decompression and writes the output to the file
* ``huff.tmp'' in the directory \a destDir.
* @param f File stream for the archive
* @param destDir The directory where the file is extracted to
* @param fInfo Info parameters for the file to be decompressed
*/
int huffExtractToFile(ifstream &f, 
                      string destDir,
                      const skSegorFile &fInfo)
{
  // Construct output filename
  string fileName(destDir);
  fileName += "huff.tmp";

  ofstream fOut(fileName.c_str());

  if (fOut)
  {

    // Read Huffman tree
    vector<skHuffNode> huffTree;
    skHuffNode hnNew;
    // Insert dummy entry
    huffTree.push_back(hnNew);

    unsigned long int cnt = 0;
    while ((!f.eof()) && (cnt < TREESIZE))
    {
      // Read entry from the file ...
      if (OK == hnNew.read(f))
      {
        // ... and insert it to the tree. 
        huffTree.push_back(hnNew);
      }
      else
      {
        fOut.close();
        return ERROR;
      }
      ++cnt;
    }

    // All entries read?
    if (cnt < TREESIZE)
    {
      fOut.close();
      return ERROR;
    }

    //
    // Read bit field and decompress it
    //
    skBitParser bParser;
    // Length of the bit field, without the last byte
    unsigned long int fieldLength = fInfo.Size - TREESIZE*6 - 1;
    int lzssAddress, bitCnt;
    int treeIndex = TREESIZE;
    // Read bits until we encounter the end of the file...
    while ((!f.eof()) && 
           //... or arrive within the last data byte ...
           (bParser.count() != fieldLength) &&
           //... or at the start of it.
           ((bParser.count() != (fieldLength - 1)) || (0 != bParser.bits())))
    {
      // Get next bit and traverse the tree...
      if (1 == bParser.getBit(f))
      {
        // ... to the right
        treeIndex = huffTree[treeIndex].Right; 
      }
      else
      {
        // ... to the left
        treeIndex = huffTree[treeIndex].Left; 
      }

      // Continue tree traversal?
      if (treeIndex <= 0)
      {
        // No, the current byte is complete
        treeIndex *= -1;
        
        // Is it a `normal' byte or a repetition length?
        if (treeIndex < 256)
        {
          // Output a `normal' byte
          writeUChar(fOut, (unsigned char) 0x00);
          writeUChar(fOut, (unsigned char) treeIndex);
        }
        else
        {
          // Write repetition length
          writeUChar(fOut, (unsigned char) 0x01);
          writeUChar(fOut, (unsigned char) (treeIndex & 0xFF));

          // Extract 12-bit address for LZSS...
          for (bitCnt = 0, lzssAddress = 0; bitCnt < 12; ++bitCnt)
            lzssAddress = (lzssAddress << 1) | bParser.getBit(f);

          // ... and output it as 16-bit integer, MSB first
          writeUChar(fOut, (unsigned char) (lzssAddress >> 8) & 0xFF);
          writeUChar(fOut, (unsigned char) (lzssAddress & 0xFF));
        }

        // Reset `tree walker' to the start index
        treeIndex = TREESIZE;
      }

    }

    // End of file encountered?
    if (f.eof())
    {
      fOut.close();
      return ERROR;
    }

    // Read the rest of the final data byte...
    int restLength;
    if (bParser.count() == (fieldLength - 1))
    {
      // The main loop was left because we are exactly
      // at the start of the final data byte, so there should
      // be no bits left...
      assert(0 == bParser.bits());

      restLength = 8;
    }
    else
      restLength = bParser.bits();

    int finalByte = 0;
    for (bitCnt = 0; bitCnt < restLength; ++bitCnt)
      finalByte = (finalByte << 1) | bParser.getBit(f);
    // Shift the left bits up to the top of the byte.
    finalByte = (finalByte << (8 - restLength));

    // Read the final byte, which equals the number of valid bits
    // in the final data byte (=finalByte)
    unsigned char c = readUChar(f);

    // Process the rest of bits
    c = c - (8 - restLength);
    // Now c should lie in the range 0-8...
    assert(c < 9);
    if (c > 8)
    {
      fOut.close();
      return ERROR;
    }

    for (bitCnt = 0; bitCnt < c; ++bitCnt)
    {
      // Get next bit and traverse the tree...
      if (0x80 == (finalByte & 0x80))
      {
        // ... to the right
        treeIndex = huffTree[treeIndex].Right; 
      }
      else
      {
        // ... to the left
        treeIndex = huffTree[treeIndex].Left; 
      }
      finalByte = (finalByte << 1) & 0xFF;

      // Continue tree traversal?
      if (treeIndex <= 0)
      {
        // No, the current byte is complete
        treeIndex *= -1;
        
        // Is it a `normal' byte or a repetition length?
        if (treeIndex < 256)
        {
          // Output a `normal' byte
          writeUChar(fOut, (unsigned char) 0x00);
          writeUChar(fOut, (unsigned char) (treeIndex & 0xFF));
        }
        else
        {
          // Return an error,since we are too close at the end
          // of the bit field to read another 12-bit value
          assert(1 == 0);
          fOut.close();
          return ERROR;
        }

        // Reset `tree walker' to the start index
        treeIndex = TREESIZE;
      }

    
    }

    fOut.close();

    return OK;
  }

  return ERROR;

}

/** Reads the file ``huff.tmp'' from the directory
* \a destDir and uncompresses it to the filename
* ``lzss.tmp'' using LZSS decoding.
* @param destDir The directory where the file is extracted to
*/
int lzssExtract(string destDir)
{
  // Construct input filename
  string fileName(destDir);
  fileName += "huff.tmp";

  ifstream f(fileName.c_str());
  if (!f)
    return ERROR;
 
  // Construct output filename
  fileName = destDir;
  fileName += "lzss.tmp";

  ofstream fOut(fileName.c_str());

  if (fOut)
  {
    unsigned char lzssBuffer[LZSSBUFLEN];
    unsigned char helpBuffer[LZSSBUFLEN];
    unsigned int helpIndex = 0;
    unsigned int bufferIndex = 0;
    unsigned int lzssAddress = 0;
    int repetitionLength = 0;
    unsigned int cnt = 0;
    unsigned char c;

    // Reset buffers
    for (bufferIndex = 0; bufferIndex < LZSSBUFLEN; ++bufferIndex)
    {
      lzssBuffer[bufferIndex] = 0x0;
      helpBuffer[bufferIndex] = 0x0;
    }
    bufferIndex = 0;

    c = readUChar(f);
    while (!f.eof())
    {
      // Is it a repetition length?
      if (0x00 == c)
      {
        // No, so write value to the file and the LZSS buffer
        c = readUChar(f);
        writeUChar(fOut, c);

        lzssBuffer[bufferIndex] = c;
        bufferIndex = (bufferIndex + 1) % LZSSBUFLEN;
      }
      else
      {
        // We get a repetition length for the LZSS buffer ...
        c = readUChar(f);
        repetitionLength = (c & 0xFF);

        // ... and an offset for the buffer index.
        c = readUChar(f);
        lzssAddress = ((c & 0xFF) << 8);
        c = readUChar(f);
        lzssAddress |= (c & 0xFF);

        // Compute index to repeat string
        lzssAddress = (bufferIndex + LZSSBUFLEN - lzssAddress) % LZSSBUFLEN;

        // Write repeat string to the LZSS buffer
        while (repetitionLength > 1)
        {
          c = lzssBuffer[lzssAddress];
          writeUChar(fOut, c);
          helpBuffer[helpIndex] = c;
          helpIndex = (helpIndex + 1) % LZSSBUFLEN;
          lzssAddress = (lzssAddress + 1) % LZSSBUFLEN;
          --repetitionLength;
        }
        if (1 == repetitionLength)
        {
          helpBuffer[helpIndex] = lzssBuffer[lzssAddress];
          helpIndex = (helpIndex + 1) % LZSSBUFLEN;

          for (cnt = 0; cnt < helpIndex; ++cnt)
          {
            lzssBuffer[bufferIndex] = helpBuffer[cnt];
            bufferIndex = (bufferIndex + 1) % LZSSBUFLEN;
          }
          writeUChar(fOut, helpBuffer[cnt-1]);
          helpIndex = 0;
        }

      }

      // Read next value
      c = readUChar(f);
    }

    fOut.close();
    return OK;
  }
  else
    f.close();

  return ERROR;
}

/** Reads the file ``lzss.tmp'' from the directory
* \a destDir and uncompresses it to the filename
* specified in \a fInfo, using runlength encoding (RLE).
* @param destDir The directory where the file is extracted to
* @param fInfo Info parameters for the file to be decompressed
*/
int rleExtractFromFile(string destDir,
                       const skSegorFile &fInfo)
{
  // Construct input filename
  string fileName(destDir);
  fileName += "lzss.tmp";

  ifstream f(fileName.c_str());
  if (!f)
    return ERROR;

  // Construct output filename
  fileName = destDir;
  fileName += fInfo.Name;
  if (fInfo.Extension.length() > 0)
  {
    fileName += ".";
    fileName += fInfo.Extension;
  }

  ofstream fOut(fileName.c_str());

  if (fOut)
  {
    unsigned char c, r, i;
    c = readUChar(f);
    while (!f.eof())
    {
      if (0x90 != c)
      {
        // Normal char -> store and write to output
        r = c;
        writeUChar(fOut, c);
      }
      else
      {
        // 0x90 encountered -> read next char
        c = readUChar(f);
        if (0x00 != c)
        {
          // Repeat last character c times
          for (i = 0; i < c; ++i)
            writeUChar(fOut, r);
        }
        else
        {
          // Convert special sequence 0x9000 to a single letter
          r = 0x90;
          writeUChar(fOut, r);
        }
      }
      c = readUChar(f);
    }

    fOut.close();
    return OK;
  }
  else
    f.close();

  return ERROR;

}

/** Extracts a single file from the Segor archive.
* @param f File stream for the archive
* @param destDir The directory where the file is extracted to
* @param fInfo Info parameters for the file to be decompressed
*/
int extractFile(ifstream &f, 
                string destDir,
                const skSegorFile &fInfo)
{

  switch (fInfo.Compression)
  {
    case RUNLENGTH: rleExtractToFile(f, destDir, fInfo);
                    break;
    case COMPLEX:   huffExtractToFile(f, destDir, fInfo);
                    lzssExtract(destDir);
                    rleExtractFromFile(destDir, fInfo);
                    break;
    default:        // Unsupported compression type
                    return ERROR;
  }

  return OK;
}


#ifdef SEGORKAT_QT

/** Uncompresses the binary file \a fileName
* into the directory \a dirName.
* @param fileName Name of the compressed binary
* @param dirName Directory where all files are extracted to
* @return OK if files were uncompressed successfully, ERROR else
*/
int uncompressSegorCatalog(const QString &fileName,
                           const QString &dirName)
{
  ifstream f(fileName.latin1());
  if (f)
  {
    // Read infos for contained files
    int i = 0, res;
    vector<skSegorFile> sfList;
    skSegorFile sfNew;
    while ((!f.eof()) && (i < NUMFILES))
    {
      // Read entry from file
      res = sfNew.read(f);
      if (OK == res)
      {
        // Append entry to list
        sfList.push_back(sfNew);
      }
      else
      {
        return res;
      }
      ++i;
    }

    if (i < NUMFILES)
      return ERROR;

    // Skip dummy value
    char c;
    f.get(c);

    Q3ProgressDialog progress( qApp->translate("skUncompress", "Entpacke Dateien..."), 
                              qApp->translate("skUncompress", "Abbruch"), NUMFILES,
                              0, "progress", TRUE );
    progress.setMinimumDuration(0);
    progress.setProgress( 1 );
    qApp->processEvents();     

    // Extract files
    QString currentFilename;
    int currentFile;
    for (currentFile = 1; currentFile < NUMFILES; ++currentFile)
    {
      currentFilename = sfList[currentFile].Name.c_str();
      if (sfList[currentFile].Extension.size() > 0)
      {
        currentFilename += ".";
        currentFilename += sfList[currentFile].Extension.c_str();
      }

      progress.setLabelText(qApp->translate("skUncompress", "Entpacke ") + currentFilename);
      qApp->processEvents();     

      if (OK != extractFile(f, string(dirName.latin1()), sfList[currentFile]))
        return ERROR;

      progress.setProgress( currentFile + 1);
      qApp->processEvents();     
      if ( progress.wasCancelled() )
        break;
    }

    // Close file
    f.close();

    return OK;
  }

  return ERROR;
} 

#else // of: #ifdef SEGORKAT_QT ...

/** Uncompresses the binary file \a fileName
* into the directory \a dirName.
* @param fileName Name of the compressed binary
* @param dirName Directory where all files are extracted to
* @return OK if files were uncompressed successfully, ERROR else
*/
int uncompressSegorCatalog(const string &fileName,
                           const string &dirName)
{
  ifstream f(fileName.c_str());
  if (f)
  {
    // Read infos for contained files
    int i = 0, res;
    vector<skSegorFile> sfList;
    skSegorFile sfNew;
    while ((!f.eof()) && (i < NUMFILES))
    {
      // Read entry from file
      res = sfNew.read(f);
      if (OK == res)
      {
        // Append entry to list
        sfList.push_back(sfNew);
      }
      else
      {
        return res;
      }
      ++i;
    }

    if (i < NUMFILES)
      return ERROR;

    // Skip dummy value
    char c;
    f.get(c);

    // Extract files
    int currentFile;
    for (currentFile = 1; currentFile < NUMFILES; ++currentFile)
    {
      if (OK != extractFile(f, dirName, sfList[currentFile]))
        return ERROR;
    }

    // Close file
    f.close();

    return OK;
  }

  return ERROR;
} 

#endif // of: #ifdef SEGORKAT_QT ... #else ...


